@@ -1,4 +1,4 @@
-/* $XFree86$ */ /* -*- mode: c; c-basic-offset: 3 -*- */
+/* -*- mode: c; c-basic-offset: 3 -*- */
 /*
  * Copyright 2000 Gareth Hughes
  * All Rights Reserved.
@@ -31,6 +31,10 @@
  *   Leif Delgass <ldelgass@retinalburn.net>
  */
 
+#include <string.h>
+#include <stdio.h>
+#include <unistd.h>
+
 /* Driver data structures */
 #include "ati.h"
 #include "atibus.h"
@@ -117,7 +121,7 @@ static Bool ATIInitVisualConfigs( ScreenPtr pScreen )
       }
 
       i = 0;
-      for (db = 0; db <= 1; db++) {
+      for (db = 1; db >= 0; db--) {
 	 for ( accum = 0 ; accum <= ATI_USE_ACCUM ; accum++ ) {
 	    for ( stencil = 0 ; stencil <= ATI_USE_STENCIL ; stencil++ ) {
 	       pATIConfigPtrs[i] = &pATIConfigs[i];
@@ -198,7 +202,7 @@ static Bool ATIInitVisualConfigs( ScreenPtr pScreen )
       }
 
       i = 0;
-      for (db = 0; db <= 1; db++) {
+      for (db = 1; db >= 0; db--) {
 	 for ( accum = 0 ; accum <= ATI_USE_ACCUM ; accum++ ) {
 	    for ( stencil = 0 ; stencil <= ATI_USE_STENCIL ; stencil++ ) {
 	       pATIConfigPtrs[i] = &pATIConfigs[i];
@@ -289,9 +293,9 @@ static void ATIEnterServer( ScreenPtr pScreen )
    ScrnInfoPtr pScreenInfo = xf86Screens[pScreen->myNum];
    ATIPtr pATI = ATIPTR(pScreenInfo);
 
-   if ( pATI->directRenderingEnabled && pATI->pXAAInfo ) { 
-      pATI->pXAAInfo->NeedToSync = TRUE;
-      pATI->NeedDRISync = TRUE;
+   if ( pATI->directRenderingEnabled ) { 
+      ATIDRIMarkSyncInt(pScreenInfo);
+      ATIDRIMarkSyncExt(pScreenInfo);
    }
 }
 
@@ -329,6 +333,7 @@ static void ATIDRISwapContext( ScreenPtr pScreen,
    }
 }
 
+#ifdef USE_XAA
 static void ATIDRITransitionTo2d(ScreenPtr pScreen)
 {
    ScrnInfoPtr pScreenInfo = xf86Screens[pScreen->myNum];
@@ -396,10 +401,45 @@ static void ATIDRITransitionTo3d(ScreenPtr pScreen)
 
    pATI->have3DWindows = TRUE;
 }
+#endif /* USE_XAA */
+
+#ifdef USE_EXA
+static void ATIDRITransitionTo2d_EXA(ScreenPtr pScreen)
+{
+   ScrnInfoPtr pScreenInfo = xf86Screens[pScreen->myNum];
+   ATIPtr pATI = ATIPTR(pScreenInfo);
+   ATIDRIServerInfoPtr pATIDRIServer = pATI->pDRIServerInfo;
+
+   exaEnableDisableFBAccess(pScreen->myNum, FALSE);
+
+   pATI->pExa->offScreenBase = pATIDRIServer->backOffset;
+
+   exaEnableDisableFBAccess(pScreen->myNum, TRUE);
+
+   pATI->have3DWindows = FALSE;
+}
+
+static void ATIDRITransitionTo3d_EXA(ScreenPtr pScreen)
+{
+   ScrnInfoPtr pScreenInfo = xf86Screens[pScreen->myNum];
+   ATIPtr pATI = ATIPTR(pScreenInfo);
+   ATIDRIServerInfoPtr pATIDRIServer = pATI->pDRIServerInfo;
+
+   exaEnableDisableFBAccess(pScreen->myNum, FALSE);
+
+   pATI->pExa->offScreenBase = pATIDRIServer->textureOffset +
+			       pATIDRIServer->textureSize;
+
+   exaEnableDisableFBAccess(pScreen->myNum, TRUE);
+
+   pATI->have3DWindows = TRUE;
+}
+#endif /* USE_EXA */
 
 /* Initialize the state of the back and depth buffers. */
 static void ATIDRIInitBuffers( WindowPtr pWin, RegionPtr prgn, CARD32 indx )
 {
+#ifdef USE_XAA
    ScreenPtr   pScreen = pWin->drawable.pScreen;
    ScrnInfoPtr pScreenInfo   = xf86Screens[pScreen->myNum];
    ATIPtr pATI = ATIPTR(pScreenInfo);
@@ -450,7 +490,8 @@ static void ATIDRIInitBuffers( WindowPtr pWin, RegionPtr prgn, CARD32 indx )
 					      pbox->x2 - pbox->x1,
 					      pbox->y2 - pbox->y1);
 
-   pXAAInfo->NeedToSync = TRUE;
+   ATIDRIMarkSyncInt(pScreenInfo);
+#endif
 }
 
 /* Copy the back and depth buffers when the X server moves a window.
@@ -465,6 +506,7 @@ static void ATIDRIInitBuffers( WindowPtr pWin, RegionPtr prgn, CARD32 indx )
 static void ATIDRIMoveBuffers( WindowPtr pWin, DDXPointRec ptOldOrg,
 			       RegionPtr prgnSrc, CARD32 indx )
 {
+#ifdef USE_XAA
     ScreenPtr pScreen = pWin->drawable.pScreen;
     ScrnInfoPtr pScreenInfo = xf86Screens[pScreen->myNum];
     ATIPtr pATI = ATIPTR(pScreenInfo);
@@ -628,7 +670,8 @@ static void ATIDRIMoveBuffers( WindowPtr pWin, DDXPointRec ptOldOrg,
     DEALLOCATE_LOCAL(pptNew1);
     DEALLOCATE_LOCAL(pboxNew1);
 
-    pXAAInfo->NeedToSync = TRUE;
+    ATIDRIMarkSyncInt(pScreenInfo);
+#endif
 }
 
 /* Compute log base 2 of val. */
@@ -641,6 +684,36 @@ static int Mach64MinBits(int val)
     return bits;
 }
 
+static Bool ATIDRISetBufSize( ScreenPtr pScreen, unsigned int maxSize )
+{
+   ScrnInfoPtr pScreenInfo = xf86Screens[pScreen->myNum];
+   ATIPtr pATI = ATIPTR(pScreenInfo);
+   ATIDRIServerInfoPtr pATIDRIServer = pATI->pDRIServerInfo;
+
+   if (pATI->OptionBufferSize) {
+      if (pATI->OptionBufferSize < 1 || pATI->OptionBufferSize > maxSize  ) {
+	 xf86DrvMsg( pScreen->myNum, X_ERROR, "[drm] Illegal DMA buffers size: %d MB\n",
+		     pATI->OptionBufferSize );
+	 return FALSE;
+      }
+      if (pATI->OptionBufferSize > 2) {
+	 xf86DrvMsg( pScreen->myNum, X_WARNING, "[drm] Illegal DMA buffers size: %d MB\n",
+		     pATI->OptionBufferSize );
+	 xf86DrvMsg( pScreen->myNum, X_WARNING, "[drm] Clamping DMA buffers size to 2 MB\n");
+	 pATIDRIServer->bufferSize = 2;
+      } else {
+	 pATIDRIServer->bufferSize = pATI->OptionBufferSize;
+	 xf86DrvMsg( pScreen->myNum, X_CONFIG, "[drm] Using %d MB for DMA buffers\n",
+		     pATIDRIServer->bufferSize );
+      }
+   } else {
+      xf86DrvMsg( pScreen->myNum, X_DEFAULT, "[drm] Using %d MB for DMA buffers\n",
+		  pATIDRIServer->bufferSize );
+   }
+
+   return TRUE;
+}
+
 static Bool ATIDRISetAgpMode( ScreenPtr pScreen )
 {
    ScrnInfoPtr pScreenInfo = xf86Screens[pScreen->myNum];
@@ -678,7 +751,6 @@ static Bool ATIDRISetAgpMode( ScreenPtr pScreen )
 
    if (pATI->OptionAGPSize) {
       switch (pATI->OptionAGPSize) {
-      case 256:
       case 128:
       case  64:
       case  32:
@@ -763,27 +835,8 @@ static Bool ATIDRIAgpInit( ScreenPtr pScreen )
    xf86DrvMsg(pScreen->myNum, X_INFO,
 	      "[agp] Using %d kB for DMA descriptor ring\n", pATIDRIServer->ringSize);
 
-   if (pATI->OptionBufferSize) {
-      if (pATI->OptionBufferSize < 1 || pATI->OptionBufferSize > pATIDRIServer->agpSize  ) {
-	 xf86DrvMsg( pScreen->myNum, X_ERROR, "[agp] Illegal DMA buffers size: %d MB\n",
-		     pATI->OptionBufferSize );
-	 return FALSE;
-      }
-      if (pATI->OptionBufferSize > 2) {
-	 xf86DrvMsg( pScreen->myNum, X_WARNING, "[agp] Illegal DMA buffers size: %d MB\n",
-		     pATI->OptionBufferSize );
-	 xf86DrvMsg( pScreen->myNum, X_WARNING, "[agp] Clamping DMA buffers size to 2 MB\n",
-		     pATI->OptionBufferSize );
-	 pATIDRIServer->bufferSize = 2;
-      } else {
-	 pATIDRIServer->bufferSize = pATI->OptionBufferSize;
-	 xf86DrvMsg( pScreen->myNum, X_CONFIG, "[agp] Using %d MB for DMA buffers\n",
-		     pATIDRIServer->bufferSize );
-      }
-   } else {
-      xf86DrvMsg( pScreen->myNum, X_DEFAULT, "[agp] Using %d MB for DMA buffers\n",
-		  pATIDRIServer->bufferSize );
-   }
+   if ( !ATIDRISetBufSize( pScreen, pATIDRIServer->agpSize ) )
+      return FALSE;
 
    pATIDRIServer->agpTexSize    = pATIDRIServer->agpSize - pATIDRIServer->bufferSize;
 
@@ -814,7 +867,7 @@ static Bool ATIDRIAgpInit( ScreenPtr pScreen )
       return FALSE;
    }
    xf86DrvMsg( pScreen->myNum, X_INFO,
-	       "[agp] ring handle = 0x%08lx\n",
+	       "[agp] ring handle = 0x%08x\n",
 	       pATIDRIServer->ringHandle );
 
    if ( drmMap( pATI->drmFD, pATIDRIServer->ringHandle,
@@ -829,13 +882,13 @@ static Bool ATIDRIAgpInit( ScreenPtr pScreen )
 
    /* Map vertex buffers */
    if ( drmAddMap( pATI->drmFD, pATIDRIServer->bufferStart, pATIDRIServer->bufferMapSize,
-		   DRM_AGP, 0, &pATIDRIServer->bufferHandle ) < 0 ) {
+		   DRM_AGP, DRM_READ_ONLY, &pATIDRIServer->bufferHandle ) < 0 ) {
       xf86DrvMsg( pScreen->myNum, X_ERROR,
 		  "[agp] Could not add vertex buffers mapping\n" );
       return FALSE;
    }
    xf86DrvMsg( pScreen->myNum, X_INFO,
-	       "[agp] vertex buffers handle = 0x%08lx\n",
+	       "[agp] vertex buffers handle = 0x%08x\n",
 	       pATIDRIServer->bufferHandle );
 
    if ( drmMap( pATI->drmFD, pATIDRIServer->bufferHandle,
@@ -856,7 +909,7 @@ static Bool ATIDRIAgpInit( ScreenPtr pScreen )
       return FALSE;
    }
    xf86DrvMsg(pScreen->myNum, X_INFO,
-	      "[agp] AGP texture region handle = 0x%08lx\n",
+	      "[agp] AGP texture region handle = 0x%08x\n",
 	      pATIDRIServer->agpTexHandle);
 
    if (drmMap(pATI->drmFD, pATIDRIServer->agpTexHandle, pATIDRIServer->agpTexMapSize,
@@ -906,6 +959,55 @@ static Bool ATIDRIAgpInit( ScreenPtr pScreen )
    return TRUE;
 }
 
+static Bool ATIDRIPciInit( ScreenPtr pScreen )
+{
+   ScrnInfoPtr pScreenInfo = xf86Screens[pScreen->myNum];
+   ATIPtr pATI = ATIPTR(pScreenInfo);
+   ATIDRIServerInfoPtr pATIDRIServer = pATI->pDRIServerInfo;
+
+   pATIDRIServer->bufferSize = ATI_DEFAULT_BUFFER_SIZE;
+   pATIDRIServer->ringSize = 16; /* 16 kB ring */
+
+   if ( !ATIDRISetBufSize( pScreen, (unsigned)(-1) ) )
+      return FALSE;
+
+   /* Set size of the DMA descriptor ring */
+   pATIDRIServer->ringStart   = 0;
+   pATIDRIServer->ringMapSize = pATIDRIServer->ringSize*1024; /* ringSize is in kB */
+
+   /* Set size of the vertex buffer */
+   pATIDRIServer->bufferStart   = 0;
+   pATIDRIServer->bufferMapSize = pATIDRIServer->bufferSize*1024*1024;
+
+   /* Map DMA descriptor ring */
+   if ( drmAddMap( pATI->drmFD, 0, pATIDRIServer->ringMapSize,
+		   DRM_CONSISTENT, DRM_RESTRICTED, &pATIDRIServer->ringHandle ) < 0 ) {
+      xf86DrvMsg( pScreen->myNum, X_ERROR,
+		  "[pci] Could not add ring mapping\n" );
+      return FALSE;
+   }
+   xf86DrvMsg( pScreen->myNum, X_INFO, "[pci] ring handle = 0x%08x\n",
+	       pATIDRIServer->ringHandle );
+
+   if ( drmMap( pATI->drmFD, pATIDRIServer->ringHandle,
+		pATIDRIServer->ringMapSize, &pATIDRIServer->ringMap ) < 0 ) {
+      xf86DrvMsg( pScreen->myNum, X_ERROR,
+		  "[pci] Could not map ring\n" );
+      return FALSE;
+   }
+   xf86DrvMsg( pScreen->myNum, X_INFO,
+	       "[pci] Ring mapped at 0x%08lx\n",
+	       (unsigned long)pATIDRIServer->ringMap );
+
+   /* Disable AGP for ForcePCIMode */
+   if ( pATI->BusType != ATI_BUS_PCI ) {
+       outm( AGP_BASE, 0 );
+       outm( AGP_CNTL, 0 );
+   }
+
+   return TRUE;
+}
+
 /* Add a map for the MMIO registers that will be accessed by any
  * DRI-based clients.
  */
@@ -925,7 +1027,7 @@ static Bool ATIDRIMapInit( ScreenPtr pScreen )
       return FALSE;
    }
    xf86DrvMsg( pScreen->myNum, X_INFO,
-	       "[drm] register handle = 0x%08lx\n",
+	       "[drm] register handle = 0x%08x\n",
 	       pATIDRIServer->regsHandle );
 
    return TRUE;
@@ -982,9 +1084,9 @@ static Bool ATIDRIAddBuffers( ScreenPtr pScreen )
    /* Initialize vertex buffers */
    if ( pATIDRIServer->IsPCI ) {
       pATIDRIServer->numBuffers = drmAddBufs( pATI->drmFD,
-					      (pATIDRIServer->bufferSize*1024*1024)/MACH64_BUFFER_SIZE,
+					      pATIDRIServer->bufferMapSize/MACH64_BUFFER_SIZE,
 					      MACH64_BUFFER_SIZE,
-					      0,
+					      DRM_PCI_BUFFER_RO,
 					      0 );
    } else {
       pATIDRIServer->numBuffers = drmAddBufs( pATI->drmFD,
@@ -1020,7 +1122,7 @@ static Bool ATIDRIMapBuffers( ScreenPtr pScreen )
    xf86DrvMsg( pScreen->myNum, X_INFO,
 	       "[drm] Mapped %d DMA buffers at 0x%08lx\n",
 	       pATIDRIServer->drmBuffers->count,
-	       pATIDRIServer->drmBuffers->list->address );
+	       (unsigned long)pATIDRIServer->drmBuffers->list->address );
 
    return TRUE;
 }
@@ -1058,8 +1160,7 @@ static Bool ATIDRIIrqInit( ScreenPtr pScreen )
 		    pATI->irq);
       else {
 	 xf86DrvMsg(pScreenInfo->scrnIndex, X_INFO,
-		    "[drm] Falling back to irq-free operation\n",
-		    pATI->irq);
+		    "[drm] Falling back to irq-free operation\n");
 	 return FALSE;
       }
    }
@@ -1087,7 +1188,6 @@ Bool ATIDRIScreenInit( ScreenPtr pScreen )
     * for known symbols in each module.
     */
    if ( !xf86LoaderCheckSymbol("GlxSetVisualConfigs") ) return FALSE;
-   if ( !xf86LoaderCheckSymbol("DRIScreenInit") ) return FALSE;
    if ( !xf86LoaderCheckSymbol("drmAvailable") ) return FALSE;
    if ( !xf86LoaderCheckSymbol("DRIQueryVersion") ) {
       xf86DrvMsg( pScreen->myNum, X_ERROR,
@@ -1097,13 +1197,13 @@ Bool ATIDRIScreenInit( ScreenPtr pScreen )
 
    /* Check the DRI version */
    DRIQueryVersion( &major, &minor, &patch );
-   if ( major != DRIINFO_MAJOR_VERSION || minor < DRIINFO_MINOR_VERSION ) {
+   if ( major != DRIINFO_MAJOR_VERSION || minor < 0 ) {
       xf86DrvMsg( pScreen->myNum, X_ERROR,
 		  "[dri] ATIDRIScreenInit failed because of a version mismatch.\n"
 		  "[dri] libdri version is %d.%d.%d but version %d.%d.x is needed.\n"
 		  "[dri] Disabling the DRI.\n",
 		  major, minor, patch,
-                  DRIINFO_MAJOR_VERSION, DRIINFO_MINOR_VERSION );
+                  DRIINFO_MAJOR_VERSION, 0 );
       return FALSE;
    }
 
@@ -1150,7 +1250,7 @@ Bool ATIDRIScreenInit( ScreenPtr pScreen )
    pDRIInfo->ddxDriverMajorVersion = ATI_VERSION_MAJOR;
    pDRIInfo->ddxDriverMinorVersion = ATI_VERSION_MINOR;
    pDRIInfo->ddxDriverPatchVersion = ATI_VERSION_PATCH;
-   pDRIInfo->frameBufferPhysicalAddress = pATI->LinearBase;
+   pDRIInfo->frameBufferPhysicalAddress = (void *)pATI->LinearBase;
    pDRIInfo->frameBufferSize = pATI->LinearSize;
    pDRIInfo->frameBufferStride = (pScreenInfo->displayWidth *
 				  pATI->FBBytesPerPixel);
@@ -1169,9 +1269,10 @@ Bool ATIDRIScreenInit( ScreenPtr pScreen )
       ErrorF( "[dri] Data does not fit in SAREA\n" );
       return FALSE;
    }
-   xf86DrvMsg( pScreenInfo->scrnIndex, X_INFO, "[drm] SAREA %d+%d: %d\n",
-	       sizeof(XF86DRISAREARec), sizeof(ATISAREAPrivRec),
-	       sizeof(XF86DRISAREARec) + sizeof(ATISAREAPrivRec) );
+   xf86DrvMsg( pScreenInfo->scrnIndex, X_INFO, "[drm] SAREA %u+%u: %u\n",
+	       (unsigned)sizeof(XF86DRISAREARec),
+	       (unsigned)sizeof(ATISAREAPrivRec),
+	       (unsigned)(sizeof(XF86DRISAREARec) + sizeof(ATISAREAPrivRec)) );
    pDRIInfo->SAREASize = SAREA_MAX;
 
    pATIDRI = (ATIDRIPtr) xnfcalloc( sizeof(ATIDRIRec), 1 );
@@ -1204,8 +1305,18 @@ Bool ATIDRIScreenInit( ScreenPtr pScreen )
    pDRIInfo->SwapContext	= ATIDRISwapContext;
    pDRIInfo->InitBuffers	= ATIDRIInitBuffers;
    pDRIInfo->MoveBuffers	= ATIDRIMoveBuffers;
-   pDRIInfo->TransitionTo2d     = ATIDRITransitionTo2d;
-   pDRIInfo->TransitionTo3d     = ATIDRITransitionTo3d;
+#ifdef USE_XAA
+   if (!pATI->useEXA) {
+      pDRIInfo->TransitionTo2d  = ATIDRITransitionTo2d;
+      pDRIInfo->TransitionTo3d  = ATIDRITransitionTo3d;
+   }
+#endif /* USE_XAA */
+#ifdef USE_EXA
+   if (pATI->useEXA) {
+      pDRIInfo->TransitionTo2d  = ATIDRITransitionTo2d_EXA;
+      pDRIInfo->TransitionTo3d  = ATIDRITransitionTo3d_EXA;
+   }
+#endif /* USE_EXA */
    pDRIInfo->bufferRequests	= DRI_ALL_WINDOWS;
 
    pDRIInfo->createDummyCtx     = TRUE;
@@ -1263,12 +1374,12 @@ Bool ATIDRIScreenInit( ScreenPtr pScreen )
    /* Check the mach64 DRM version */
    version = drmGetVersion( pATI->drmFD );
    if ( version ) {
-      if ( version->version_major != 1 ||
+      if ( version->version_major != 2 ||
 	   version->version_minor < 0 ) {
 	 /* Incompatible DRM version */
 	 xf86DrvMsg( pScreen->myNum, X_ERROR,
 		     "[dri] ATIDRIScreenInit failed because of a version mismatch.\n"
-		     "[dri] mach64.o kernel module version is %d.%d.%d, but version 1.0 or greater is needed.\n"
+		     "[dri] mach64.o kernel module version is %d.%d.%d, but version 2.x is needed (with 2.x >= 2.0).\n"
 		     "[dri] Disabling DRI.\n",
 		     version->version_major,
 		     version->version_minor,
@@ -1297,49 +1408,22 @@ Bool ATIDRIScreenInit( ScreenPtr pScreen )
    pATIDRIServer->IsPCI = (pATI->BusType == ATI_BUS_PCI || pATI->OptionIsPCI) ? TRUE : FALSE;
 
    if ( pATI->BusType != ATI_BUS_PCI && pATI->OptionIsPCI ) {
-       outm( AGP_BASE, 0 );
-       outm( AGP_CNTL, 0 );
        xf86DrvMsg(pScreen->myNum, X_CONFIG, "[dri] Forcing PCI mode\n");
    }
 
-   /* Check buffer size option for PCI, since it won't be done in ATIDRIAgpInit */
-   if ( pATIDRIServer->IsPCI) {
-      pATIDRIServer->bufferSize = ATI_DEFAULT_BUFFER_SIZE;
-      if (pATI->OptionBufferSize) {
-      	 if (pATI->OptionBufferSize < 1) {
-	    xf86DrvMsg( pScreen->myNum, X_ERROR, "[pci] Illegal DMA buffers size: %d MB\n",
-			pATI->OptionBufferSize );
-	    ATIDRICloseScreen( pScreen );
-	    return FALSE;
-	 }
-	 if (pATI->OptionBufferSize > 2) {
-	    xf86DrvMsg( pScreen->myNum, X_WARNING, "[pci] Illegal DMA buffers size: %d MB\n",
-			pATI->OptionBufferSize );
-	    xf86DrvMsg( pScreen->myNum, X_WARNING, "[pci] Clamping DMA buffers size to 2 MB\n",
-			pATI->OptionBufferSize );
-	    pATIDRIServer->bufferSize = 2;
-	 } else {
-	    pATIDRIServer->bufferSize = pATI->OptionBufferSize;
-	    xf86DrvMsg( pScreen->myNum, X_CONFIG, "[pci] Using %d MB DMA buffer size\n",
-			pATIDRIServer->bufferSize );
-	 }
-      } else {
-	 xf86DrvMsg( pScreen->myNum, X_DEFAULT, "[pci] Using %d MB DMA buffer size\n",
-		     pATIDRIServer->bufferSize );
-      }
-   }
-
    /* Initialize AGP */
    if ( !pATIDRIServer->IsPCI && !ATIDRIAgpInit( pScreen ) ) {
       pATIDRIServer->IsPCI = TRUE;
-      if ( pATI->BusType != ATI_BUS_PCI ) {
-	 outm( AGP_BASE, 0 );
-	 outm( AGP_CNTL, 0 );
-      }
       xf86DrvMsg( pScreen->myNum, X_WARNING, "[agp] AGP failed to initialize -- falling back to PCI mode.\n" );
       xf86DrvMsg( pScreen->myNum, X_WARNING, "[agp] Make sure you have the agpgart kernel module loaded.\n" );
    }
 
+   /* Initialize PCI */
+   if ( pATIDRIServer->IsPCI && !ATIDRIPciInit( pScreen ) ) {
+      ATIDRICloseScreen( pScreen );
+      return FALSE;
+   }
+
    if ( !ATIDRIMapInit( pScreen ) ) {
       ATIDRICloseScreen( pScreen );
       return FALSE;
@@ -1461,8 +1545,6 @@ void ATIDRIResume( ScreenPtr pScreen )
    ATIPtr pATI = ATIPTR(pScreenInfo);
    ATIDRIServerInfoPtr pATIDRIServer = pATI->pDRIServerInfo;
 
-   int ret;
-
    xf86DrvMsg( pScreen->myNum, X_INFO,
 		 "[RESUME] Attempting to re-init Mach64 hardware.\n");
 
@@ -1517,6 +1599,10 @@ void ATIDRICloseScreen( ScreenPtr pScreen )
       drmUnmap( pATIDRIServer->bufferMap, pATIDRIServer->bufferMapSize );
       pATIDRIServer->bufferMap = NULL;
    }
+   if ( pATIDRIServer->ringMap ) {
+      drmUnmap( pATIDRIServer->ringMap, pATIDRIServer->ringMapSize );
+      pATIDRIServer->ringMap = NULL;
+   }
    if ( pATIDRIServer->agpHandle ) {
       drmAgpUnbind( pATI->drmFD, pATIDRIServer->agpHandle );
       drmAgpFree( pATI->drmFD, pATIDRIServer->agpHandle );
@@ -1524,6 +1610,12 @@ void ATIDRICloseScreen( ScreenPtr pScreen )
       drmAgpRelease( pATI->drmFD );
    }
 
+   /* De-allocate all PCI resources */
+   if ( pATIDRIServer->IsPCI && pATIDRIServer->ringHandle ) {
+      drmRmMap( pATI->drmFD, pATIDRIServer->ringHandle );
+      pATIDRIServer->ringHandle = 0;
+   }
+
    /* De-allocate all DRI resources */
    DRICloseScreen( pScreen );
 
